from bbot.modules.base import BaseModule


class robots(BaseModule):
    watched_events = ["URL"]
    produced_events = ["URL_UNVERIFIED"]
    flags = ["active", "safe", "web-basic"]
    meta = {"description": "Look for and parse robots.txt"}

    options = {"include_sitemap": False, "include_allow": True, "include_disallow": True}
    options_desc = {
        "include_sitemap": "Include 'sitemap' entries",
        "include_allow": "Include 'Allow' Entries",
        "include_disallow": "Include 'Disallow' Entries",
    }

    in_scope_only = True

    def setup(self):
        self.scanned_hosts = set()
        return True

    def handle_event(self, event):
        parsed_host = event.parsed
        host = f"{parsed_host.scheme}://{parsed_host.netloc}/"
        host_hash = hash(host)
        if host_hash in self.scanned_hosts:
            self.debug(f"Host {host} was already scanned, exiting")
            return
        else:
            self.scanned_hosts.add(host_hash)

        result = None
        url = f"{host}robots.txt"
        result = self.helpers.request(url)
        if result:
            body = result.text

            if body:
                lines = body.split("\n")
                for l in lines:
                    if len(l) > 0:
                        split_l = l.split(": ")
                        if (split_l[0].lower() == "allow" and self.config.get("include_allow") == True) or (
                            split_l[0].lower() == "disallow" and self.config.get("include_disallow") == True
                        ):
                            unverified_url = f"{host}{split_l[1].lstrip('/')}".replace(
                                "*", self.helpers.rand_string(4)
                            )

                        elif split_l[0].lower() == "sitemap" and self.config.get("include_sitemap") == True:
                            unverified_url = split_l[1]
                        else:
                            continue

                        self.emit_event(unverified_url, "URL_UNVERIFIED", source=event)



import dns.zone
import dns.query

from bbot.modules.base import BaseModule


class dnszonetransfer(BaseModule):
    flags = ["subdomain-enum", "active", "safe"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {"description": "Attempt DNS zone transfers"}
    options = {"timeout": 10}
    options_desc = {"timeout": "Max seconds to wait before timing out"}
    max_event_handlers = 5
    suppress_dupes = False

    def setup(self):
        self.timeout = self.config.get("timeout", 10)
        return True

    def filter_event(self, event):
        if any([x in event.tags for x in ("ns_record", "soa_record")]):
            return True
        return False

    def handle_event(self, event):
        domain = event.data
        self.debug("Finding nameservers with NS/SOA query")
        nameservers = list(self.helpers.resolve(event.data, type=("NS", "SOA")))
        nameserver_ips = set()
        for n in nameservers:
            nameserver_ips.update(self.helpers.resolve(n))
        self.debug(f"Found {len(nameservers):} nameservers for domain {domain}")
        for nameserver in nameserver_ips:
            if self.scan.stopping:
                break
            try:
                self.debug(f"Attempting zone transfer against {nameserver} for domain {domain}")
                xfr_answer = dns.query.xfr(nameserver, domain, timeout=self.timeout, lifetime=self.timeout)
                zone = dns.zone.from_xfr(xfr_answer)
            except Exception as e:
                self.debug(f"Error retrieving zone: {e}")
                continue
            self.hugesuccess(f"Successful zone transfer against {nameserver} for domain {domain}!")
            finding_description = f"Successful DNS zone transfer against {nameserver} for {domain}"
            self.emit_event({"host": str(event.host), "description": finding_description}, "FINDING", source=event)
            for name, ttl, rdata in zone.iterate_rdatas():
                if str(name) == "@":
                    parent_data = domain
                else:
                    parent_data = f"{name}.{domain}"
                parent_event = self.make_event(parent_data, "DNS_NAME", event)
                if not parent_event or parent_event == event:
                    parent_event = event
                else:
                    self.emit_event(parent_event)
                for rdtype, t in self.helpers.dns.extract_targets(rdata):
                    if not self.helpers.is_ip(t):
                        t = f"{t}.{domain}"
                    module = self.helpers.dns._get_dummy_module(rdtype)
                    child_event = self.scan.make_event(t, "DNS_NAME", parent_event, module=module)
                    self.emit_event(child_event)
            else:
                self.debug(f"No data returned by {nameserver} for domain {domain}")



import csv
from contextlib import suppress

from bbot.modules.output.base import BaseOutputModule


class CSV(BaseOutputModule):
    watched_events = ["*"]
    meta = {"description": "Output to CSV"}
    options = {"output_file": ""}
    options_desc = {"output_file": "Output to CSV file"}
    emit_graph_trail = False

    header_row = ["Event type", "Event data", "IP Address", "Source Module", "Scope Distance", "Event Tags"]
    filename = "output.csv"

    def setup(self):
        self._writer = None
        self._prep_output_dir(self.filename)
        return True

    @property
    def writer(self):
        if self._writer is None:
            self._writer = csv.writer(self.file)
            self._writer.writerow(self.header_row)
        return self._writer

    @property
    def file(self):
        if self._file is None:
            self._file = open(self.output_file, mode="a", newline="")
        return self._file

    def writerow(self, row):
        self.writer.writerow(row)
        self.file.flush()

    def handle_event(self, event):
        self.writerow(
            [
                getattr(event, "type", ""),
                getattr(event, "data", ""),
                ",".join(str(x) for x in getattr(event, "resolved_hosts", set()) if self.helpers.is_ip(x)),
                str(getattr(event, "module", "")),
                str(getattr(event, "scope_distance", "")),
                ",".join(sorted(list(getattr(event, "tags", [])))),
            ]
        )

    def cleanup(self):
        if getattr(self, "_file", None) is not None:
            with suppress(Exception):
                self.file.close()

    def report(self):
        if self._file is not None:
            self.info(f"Saved CSV output to {self.output_file}")



import logging
import threading
import traceback
from sys import exc_info
from pathlib import Path
import concurrent.futures
from datetime import datetime
from omegaconf import OmegaConf
from contextlib import suppress
from collections import OrderedDict, deque

from bbot import config as bbot_config

from .stats import ScanStats
from .target import ScanTarget
from .manager import ScanManager
from .dispatcher import Dispatcher
from bbot.modules import module_loader
from bbot.core.event import make_event
from bbot.core.helpers.misc import sha1, rand_string
from bbot.core.helpers.helper import ConfigAwareHelper
from bbot.core.logger import init_logging, get_log_level
from bbot.core.helpers.names_generator import random_name
from bbot.core.configurator.environ import prepare_environment
from bbot.core.helpers.threadpool import ThreadPoolWrapper, BBOTThreadPoolExecutor
from bbot.core.errors import BBOTError, ScanError, ScanCancelledError, ValidationError

log = logging.getLogger("bbot.scanner")

init_logging()


class Scanner:
    _status_codes = {
        "NOT_STARTED": 0,
        "STARTING": 1,
        "RUNNING": 2,
        "FINISHING": 3,
        "CLEANING_UP": 4,
        "ABORTING": 5,
        "ABORTED": 6,
        "FAILED": 7,
        "FINISHED": 8,
    }

    def __init__(
        self,
        *targets,
        whitelist=None,
        blacklist=None,
        scan_id=None,
        name=None,
        modules=None,
        output_modules=None,
        config=None,
        dispatcher=None,
        strict_scope=False,
        force_start=False,
    ):
        if modules is None:
            modules = []
        if output_modules is None:
            output_modules = ["python"]

        if config is None:
            config = OmegaConf.create({})
        else:
            config = OmegaConf.create(config)
        self.config = OmegaConf.merge(bbot_config, config)
        prepare_environment(self.config)

        if name is None:
            self.name = random_name()
        else:
            self.name = str(name)
        self.strict_scope = strict_scope
        self.force_start = force_start

        if scan_id is not None:
            self.id = str(scan_id)
        else:
            self.id = f"SCAN:{sha1(rand_string(20)).hexdigest()}"
        self._status = "NOT_STARTED"
        self._status_code = 0

        # Set up thread pools
        max_workers = max(1, self.config.get("max_threads", 25))
        # Shared thread pool, for module use
        self._thread_pool = BBOTThreadPoolExecutor(max_workers=max_workers)
        # Event thread pool, for event emission
        self._event_thread_pool = ThreadPoolWrapper(
            BBOTThreadPoolExecutor(max_workers=max_workers * 2), qsize=max_workers
        )
        # Internal thread pool, for handle_event(), module setup, cleanup callbacks, etc.
        self._internal_thread_pool = ThreadPoolWrapper(BBOTThreadPoolExecutor(max_workers=max_workers))
        self.process_pool = ThreadPoolWrapper(concurrent.futures.ProcessPoolExecutor())

        self.helpers = ConfigAwareHelper(config=self.config, scan=self)
        output_dir = self.config.get("output_dir", "")
        if output_dir:
            self.home = Path(output_dir).resolve() / self.name
        else:
            self.home = self.helpers.bbot_home / "scans" / self.name

        self.target = ScanTarget(self, *targets, strict_scope=strict_scope)

        self.modules = OrderedDict({})
        self._scan_modules = modules
        self._internal_modules = list(self._internal_modules())
        self._output_modules = output_modules
        self._modules_loaded = False

        if not whitelist:
            self.whitelist = self.target.copy()
        else:
            self.whitelist = ScanTarget(self, *whitelist, strict_scope=strict_scope)
        if not blacklist:
            blacklist = []
        self.blacklist = ScanTarget(self, *blacklist)

        if dispatcher is None:
            self.dispatcher = Dispatcher()
        else:
            self.dispatcher = dispatcher
        self.dispatcher.set_scan(self)

        self.manager = ScanManager(self)
        self.stats = ScanStats(self)

        # scope distance
        self.scope_search_distance = max(0, int(self.config.get("scope_search_distance", 1)))
        self.dns_search_distance = max(
            self.scope_search_distance, int(self.config.get("scope_dns_search_distance", 3))
        )
        self.scope_report_distance = int(self.config.get("scope_report_distance", 1))

        # custom HTTP headers warning
        self.custom_http_headers = self.config.get("http_headers", {})
        if self.custom_http_headers:
            self.warning(
                "You have enabled custom HTTP headers. These will be attached to all in-scope requests and all requests made by httpx."
            )

        self._prepped = False
        self._cleanedup = False

    def prep(self):
        self.helpers.mkdir(self.home)
        if not self._prepped:
            start_msg = f"Scan with {len(self._scan_modules):,} modules seeded with {len(self.target):,} targets"
            details = []
            if self.whitelist != self.target:
                details.append(f"{len(self.whitelist):,} in whitelist")
            if self.blacklist:
                details.append(f"{len(self.blacklist):,} in blacklist")
            if details:
                start_msg += f" ({', '.join(details)})"
            self.hugeinfo(start_msg)

            self.load_modules()

            self.info(f"Setting up modules...")
            self.setup_modules()

            self.success(f"Setup succeeded for {len(self.modules):,} modules.")
            self._prepped = True

    def start_without_generator(self):
        deque(self.start(), maxlen=0)

    def start(self):
        self.prep()

        failed = True

        if not self.target:
            self.warning(f"No scan targets specified")

        scan_start_time = datetime.now()
        try:
            self.status = "STARTING"

            if not self.modules:
                self.error(f"No modules loaded")
                self.status = "FAILED"
                return
            else:
                self.hugesuccess(f"Starting scan {self.name}")

            if self.stopping:
                return

            # distribute seed events
            self.manager.init_events()

            if self.stopping:
                return

            self.status = "RUNNING"
            self.start_modules()
            self.verbose(f"{len(self.modules):,} modules started")

            if self.stopping:
                return

            yield from self.manager.loop_until_finished()
            failed = False

        except KeyboardInterrupt:
            self.stop()
            failed = False

        except ScanCancelledError:
            self.debug("Scan cancelled")

        except ScanError as e:
            self.error(f"{e}")

        except BBOTError as e:
            self.critical(f"Error during scan: {e}")
            self.trace()

        except Exception:
            self.critical(f"Unexpected error during scan:\n{traceback.format_exc()}")

        finally:
            self.cleanup()
            self.shutdown_threadpools(wait=True)

            log_fn = self.hugesuccess
            if self.status == "ABORTING":
                self.status = "ABORTED"
                log_fn = self.hugewarning
            elif failed:
                self.status = "FAILED"
                log_fn = self.critical
            else:
                self.status = "FINISHED"

            scan_run_time = datetime.now() - scan_start_time
            scan_run_time = self.helpers.human_timedelta(scan_run_time)
            log_fn(f"Scan {self.name} completed in {scan_run_time} with status {self.status}")

            self.dispatcher.on_finish(self)

    def start_modules(self):
        self.verbose(f"Starting module threads")
        for module_name, module in self.modules.items():
            module.start()

    def setup_modules(self, remove_failed=True):
        self.load_modules()
        self.verbose(f"Setting up modules")
        hard_failed = []
        soft_failed = []
        setup_futures = dict()

        for module_name, module in self.modules.items():
            future = self._internal_thread_pool.submit_task(module._setup)
            setup_futures[future] = module_name
        for future in self.helpers.as_completed(setup_futures):
            module_name = setup_futures[future]
            status, msg = future.result()
            if status == True:
                self.debug(f"Setup succeeded for {module_name} ({msg})")
            elif status == False:
                self.error(f"Setup hard-failed for {module_name}: {msg}")
                self.modules[module_name].set_error_state()
                hard_failed.append(module_name)
            else:
                self.warning(f"Setup soft-failed for {module_name}: {msg}")
                soft_failed.append(module_name)
            if not status and remove_failed:
                self.modules.pop(module_name)

        num_output_modules = len([m for m in self.modules.values() if m._type == "output"])
        if num_output_modules < 1:
            raise ScanError("Failed to load output modules. Aborting.")
        total_failed = len(hard_failed + soft_failed)
        if hard_failed:
            msg = f"Setup hard-failed for {len(hard_failed):,} modules ({','.join(hard_failed)})"
            self.fail_setup(msg)
        elif total_failed > 0:
            self.warning(f"Setup failed for {total_failed:,} modules")

    def stop(self, wait=False):
        if self.status != "ABORTING":
            self.status = "ABORTING"
            self.hugewarning(f"Aborting scan")
            self.helpers.kill_children()
            self.shutdown_threadpools(wait=False)
            self.helpers.kill_children()

    def shutdown_threadpools(self, wait=True):
        pools = [
            self.process_pool,
            self._internal_thread_pool,
            self.helpers.dns._thread_pool,
            self._event_thread_pool,
            self._thread_pool,
        ]
        self.debug(f"Shutting down thread pools with wait={wait}")
        threads = []
        for pool in pools:
            t = threading.Thread(target=pool.shutdown, kwargs={"wait": False, "cancel_futures": True}, daemon=True)
            t.start()
            threads.append(t)
        if wait:
            for t in threads:
                t.join()
        if wait:
            for pool in pools:
                pool.shutdown(wait=True)
        self.debug("Finished shutting down thread pools")

    def cleanup(self):
        # clean up modules
        self.status = "CLEANING_UP"
        for mod in self.modules.values():
            mod._cleanup()
        if not self._cleanedup:
            self._cleanedup = True
            with suppress(Exception):
                self.home.rmdir()
            self.helpers.clean_old_scans()

    def in_scope(self, e):
        """
        Checks whitelist and blacklist, also taking scope_distance into account
        """
        try:
            e = make_event(e, dummy=True)
        except ValidationError:
            return False
        in_scope = e.scope_distance == 0 or self.whitelisted(e)
        return in_scope and not self.blacklisted(e)

    def blacklisted(self, e):
        e = make_event(e, dummy=True)
        return e in self.blacklist

    def whitelisted(self, e):
        e = make_event(e, dummy=True)
        return e in self.whitelist

    @property
    def word_cloud(self):
        return self.helpers.word_cloud

    @property
    def stopping(self):
        return not self.running

    @property
    def running(self):
        return 0 < self._status_code < 4

    @property
    def aborting(self):
        return 5 <= self._status_code <= 6

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        """
        Block setting after status has been aborted
        """
        status = str(status).strip().upper()
        if status in self._status_codes:
            if self.status == "ABORTING" and not status == "ABORTED":
                self.debug(f'Attempt to set invalid status "{status}" on aborted scan')
            else:
                self._status = status
                self._status_code = self._status_codes[status]
                self.dispatcher.on_status(self._status, self.id)
        else:
            self.debug(f'Attempt to set invalid status "{status}" on scan')

    @property
    def status_detailed(self):
        event_threadpool_tasks = self._event_thread_pool.num_tasks
        internal_tasks = self._internal_thread_pool.num_tasks
        process_tasks = self.process_pool.num_tasks
        total_tasks = event_threadpool_tasks + internal_tasks + process_tasks
        status = {
            "queued_tasks": {
                "internal": internal_tasks,
                "process": process_tasks,
                "event": event_threadpool_tasks,
                "total": total_tasks,
            },
        }
        return status

    def make_event(self, *args, **kwargs):
        kwargs["scan"] = self
        event = make_event(*args, **kwargs)
        return event

    @property
    def log(self):
        if self._log is None:
            self._log = logging.getLogger(f"bbot.agent.scanner")
        return self._log

    @property
    def root_event(self):
        root_event = self.make_event(data=f"{self.name} ({self.id})", event_type="SCAN", dummy=True)
        root_event._id = self.id
        root_event.scope_distance = 0
        root_event._resolved.set()
        root_event.source = root_event
        root_event.module = self.helpers._make_dummy_module(name="TARGET", _type="TARGET")
        return root_event

    @property
    def useragent(self):
        return self.config.get("user_agent", "BBOT")

    @property
    def json(self):
        j = dict()
        for i in ("id", "name"):
            v = getattr(self, i, "")
            if v:
                j.update({i: v})
        if self.target:
            j.update({"targets": [str(e.data) for e in self.target]})
        if self.whitelist:
            j.update({"whitelist": [str(e.data) for e in self.whitelist]})
        if self.blacklist:
            j.update({"blacklist": [str(e.data) for e in self.blacklist]})
        if self.modules:
            j.update({"modules": [str(m) for m in self.modules]})
        return j

    def debug(self, *args, **kwargs):
        log.debug(*args, extra={"scan_id": self.id}, **kwargs)

    def verbose(self, *args, **kwargs):
        log.verbose(*args, extra={"scan_id": self.id}, **kwargs)

    def hugeverbose(self, *args, **kwargs):
        log.hugeverbose(*args, extra={"scan_id": self.id}, **kwargs)

    def info(self, *args, **kwargs):
        log.info(*args, extra={"scan_id": self.id}, **kwargs)

    def hugeinfo(self, *args, **kwargs):
        log.hugeinfo(*args, extra={"scan_id": self.id}, **kwargs)

    def success(self, *args, **kwargs):
        log.success(*args, extra={"scan_id": self.id}, **kwargs)

    def hugesuccess(self, *args, **kwargs):
        log.hugesuccess(*args, extra={"scan_id": self.id}, **kwargs)

    def warning(self, *args, **kwargs):
        log.warning(*args, extra={"scan_id": self.id}, **kwargs)
        self.trace()

    def hugewarning(self, *args, **kwargs):
        log.hugewarning(*args, extra={"scan_id": self.id}, **kwargs)
        self.trace()

    def error(self, *args, **kwargs):
        log.error(*args, extra={"scan_id": self.id}, **kwargs)
        self.trace()

    def trace(self):
        e_type, e_val, e_traceback = exc_info()
        if e_type is not None:
            log.trace(traceback.format_exc())

    def critical(self, *args, **kwargs):
        log.critical(*args, extra={"scan_id": self.id}, **kwargs)

    def _internal_modules(self):
        for modname in module_loader.preloaded(type="internal"):
            if self.config.get(modname, True):
                yield modname

    def load_modules(self):
        if not self._modules_loaded:
            all_modules = list(set(self._scan_modules + self._output_modules + self._internal_modules))
            if not all_modules:
                self.warning(f"No modules to load")
                return

            if not self._scan_modules:
                self.warning(f"No scan modules to load")

            # install module dependencies
            succeeded, failed = self.helpers.depsinstaller.install(
                *self._scan_modules, *self._output_modules, *self._internal_modules
            )
            if failed:
                msg = f"Failed to install dependencies for {len(failed):,} modules: {','.join(failed)}"
                self.fail_setup(msg)
            modules = sorted([m for m in self._scan_modules if m in succeeded])
            output_modules = sorted([m for m in self._output_modules if m in succeeded])
            internal_modules = sorted([m for m in self._internal_modules if m in succeeded])

            # Load scan modules
            self.verbose(f"Loading {len(modules):,} scan modules: {','.join(modules)}")
            loaded_modules, failed = self._load_modules(modules)
            self.modules.update(loaded_modules)
            if len(failed) > 0:
                msg = f"Failed to load {len(failed):,} scan modules: {','.join(failed)}"
                self.fail_setup(msg)
            if loaded_modules:
                self.info(
                    f"Loaded {len(loaded_modules):,}/{len(self._scan_modules):,} scan modules ({','.join(loaded_modules)})"
                )

            # Load internal modules
            self.verbose(f"Loading {len(internal_modules):,} internal modules: {','.join(internal_modules)}")
            loaded_internal_modules, failed_internal = self._load_modules(internal_modules)
            self.modules.update(loaded_internal_modules)
            if len(failed_internal) > 0:
                msg = f"Failed to load {len(loaded_internal_modules):,} internal modules: {','.join(loaded_internal_modules)}"
                self.fail_setup(msg)
            if loaded_internal_modules:
                self.info(
                    f"Loaded {len(loaded_internal_modules):,}/{len(self._internal_modules):,} internal modules ({','.join(loaded_internal_modules)})"
                )

            # Load output modules
            self.verbose(f"Loading {len(output_modules):,} output modules: {','.join(output_modules)}")
            loaded_output_modules, failed_output = self._load_modules(output_modules)
            self.modules.update(loaded_output_modules)
            if len(failed_output) > 0:
                msg = f"Failed to load {len(failed_output):,} output modules: {','.join(failed_output)}"
                self.fail_setup(msg)
            if loaded_output_modules:
                self.info(
                    f"Loaded {len(loaded_output_modules):,}/{len(self._output_modules):,} output modules, ({','.join(loaded_output_modules)})"
                )

            self.modules = OrderedDict(sorted(self.modules.items(), key=lambda x: getattr(x[-1], "_priority", 0)))
            self._modules_loaded = True

    def fail_setup(self, msg):
        msg = str(msg)
        if not self.force_start:
            msg += " (--force to run module anyway)"
        if self.force_start:
            self.error(msg)
        else:
            raise ScanError(msg)

    @property
    def log_level(self):
        return get_log_level()

    def _load_modules(self, modules):
        modules = [str(m) for m in modules]
        loaded_modules = {}
        failed = set()
        for module_name, module_class in module_loader.load_modules(modules).items():
            if module_class:
                try:
                    loaded_modules[module_name] = module_class(self)
                    self.verbose(f'Loaded module "{module_name}"')
                    continue
                except Exception:
                    self.warning(f"Failed to load module {module_class}")
            else:
                self.warning(f'Failed to load unknown module "{module_name}"')
            failed.add(module_name)
        return loaded_modules, failed



from bbot.modules.bucket_aws import bucket_aws


class bucket_azure(bucket_aws):
    watched_events = ["DNS_NAME", "STORAGE_BUCKET"]
    produced_events = ["STORAGE_BUCKET", "FINDING"]
    flags = ["active", "safe", "cloud-enum"]
    meta = {"description": "Check for Azure storage blobs related to target"}
    options = {"max_threads": 10, "permutations": False}
    options_desc = {
        "max_threads": "Maximum number of threads for HTTP requests",
        "permutations": "Whether to try permutations",
    }

    cloud_helper_name = "azure"
    delimiters = ("", "-")
    base_domains = ["blob.core.windows.net"]
    # Dirbusting is required to know whether a bucket is public
    supports_open_check = False

    def check_bucket_exists(self, bucket_name, url):
        response = self.helpers.request(url)
        status_code = getattr(response, "status_code", 0)
        existent_bucket = status_code != 0
        return (existent_bucket, set())

    def check_bucket_open(self, bucket_name, url):
        response = self.helpers.request(url)
        tags = self.gen_tags_exists(response)
        status_code = getattr(response, "status_code", 404)
        content = getattr(response, "text", "")
        open_bucket = status_code == 200 and "Contents" in content
        msg = ""
        if open_bucket:
            msg = "Open storage bucket"
        return (msg, tags)



from bbot.modules.base import BaseModule
from Wappalyzer import Wappalyzer, WebPage

import warnings

warnings.filterwarnings(
    "ignore",
    message="""Caught 'unbalanced parenthesis at position 119' compiling regex""",
    category=UserWarning,
)


class wappalyzer(BaseModule):
    watched_events = ["HTTP_RESPONSE"]
    produced_events = ["TECHNOLOGY"]
    flags = ["active", "safe", "web-basic"]
    meta = {
        "description": "Extract technologies from web responses",
    }
    deps_pip = ["python-Wappalyzer"]
    # accept all events regardless of scope distance
    scope_distance_modifier = None
    max_event_handlers = 5

    def setup(self):
        self.wappalyzer = Wappalyzer.latest()
        return True

    def handle_event(self, event):
        for res in self.wappalyze(event.data):
            self.emit_event(
                {"technology": res.lower(), "url": event.data["url"], "host": str(event.host)}, "TECHNOLOGY", event
            )

    def wappalyze(self, data):
        w = WebPage(url=data["url"], html=data.get("body", ""), headers=data.get("header-dict", {}))
        return self.wappalyzer.analyze(w)



import json
import threading
import websocket
from time import sleep

from bbot.modules.output.base import BaseOutputModule


class Websocket(BaseOutputModule):
    watched_events = ["*"]
    meta = {"description": "Output to websockets"}
    options = {"url": "", "token": ""}
    options_desc = {"url": "Web URL", "token": "Authorization Bearer token"}

    def setup(self):
        self.url = self.config.get("url", "")
        if not self.url:
            self.warning("Must set URL")
            return False
        kwargs = {}
        self.token = self.config.get("token", "")
        if self.token:
            kwargs.update({"header": {"Authorization": f"Bearer {self.token}"}})
        self.ws = websocket.WebSocketApp(self.url, **kwargs)
        self.started = False
        return True

    def start_websocket(self):
        if not self.started:
            self.thread = threading.Thread(target=self._start_websocket, daemon=True)
            self.thread.start()
            self.started = True

    def _start_websocket(self):
        not_keyboardinterrupt = False
        while not self.scan.stopping:
            not_keyboardinterrupt = self.ws.run_forever()
            if not not_keyboardinterrupt:
                break
            sleep(1)

    def handle_event(self, event):
        self.start_websocket()
        event_json = event.json()
        self.send(event_json)

    def send(self, message):
        while self.ws is not None:
            try:
                self.ws.send(json.dumps(message))
                break
            except Exception as e:
                self.warning(f"Error sending message: {e}, retrying")
                sleep(1)
                continue

    def cleanup(self):
        self.ws.close()
        self.ws = None



import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import RequestException

from bbot.modules.output.base import BaseOutputModule


class HTTP(BaseOutputModule):
    watched_events = ["*"]
    meta = {"description": "Send every event to a custom URL via a web request"}
    options = {
        "url": "",
        "method": "POST",
        "bearer": "",
        "username": "",
        "password": "",
        "timeout": 10,
    }
    options_desc = {
        "url": "Web URL",
        "method": "HTTP method",
        "bearer": "Authorization Bearer token",
        "username": "Username (basic auth)",
        "password": "Password (basic auth)",
        "timeout": "HTTP timeout",
    }

    def setup(self):
        self.session = requests.Session()
        if not self.config.get("url", ""):
            self.warning("Must set URL")
            return False
        if not self.config.get("method", ""):
            self.warning("Must set HTTP method")
            return False
        return True

    def handle_event(self, event):
        r = requests.Request(
            url=self.config.get("url"),
            method=self.config.get("method", "POST"),
        )
        r.headers["User-Agent"] = self.scan.useragent
        r.json = dict(event)
        username = self.config.get("username", "")
        password = self.config.get("password", "")
        if username:
            r.auth = HTTPBasicAuth(username, password)
        bearer = self.config.get("bearer", "")
        if bearer:
            r.headers["Authorization"] = f"Bearer {bearer}"
        try:
            timeout = self.config.get("timeout", 10)
            self.session.send(r.prepare(), timeout=timeout)
        except RequestException as e:
            self.warning(f"Error sending {event}: {e}")



from threading import Lock

from bbot.modules.base import BaseModule
from bbot.core.errors import NTLMError, RequestException

ntlm_discovery_endpoints = [
    "",
    "autodiscover/autodiscover.xml" "ecp/",
    "ews/",
    "ews/exchange.asmx",
    "exchange/",
    "exchweb/",
    "oab/",
    "owa/",
    "_windows/default.aspx?ReturnUrl=/",
    "abs/",
    "adfs/ls/wia",
    "adfs/services/trust/2005/windowstransport",
    "aspnet_client/",
    "autodiscover/",
    "autoupdate/",
    "certenroll/",
    "certprov/",
    "certsrv/",
    "conf/",
    "debug/",
    "deviceupdatefiles_ext/",
    "deviceupdatefiles_int/",
    "dialin/",
    "etc/",
    "groupexpansion/",
    "hybridconfig/",
    "iwa/authenticated.aspx",
    "iwa/iwa_test.aspx",
    "mcx/",
    "mcx/mcxservice.svc",
    "meet/",
    "meeting/",
    "microsoft-server-activesync/",
    "ocsp/",
    "persistentchat/",
    "phoneconferencing/",
    "powershell/",
    "public/",
    "reach/sip.svc",
    "remoteDesktopGateway/",
    "requesthandler/",
    "requesthandlerext/",
    "rgs/",
    "rgsclients/",
    "rpc/",
    "rpcwithcert/",
    "scheduler/",
    "ucwa/",
    "unifiedmessaging/",
    "webticket/",
    "webticket/webticketservice.svc",
]

NTLM_test_header = {"Authorization": "NTLM TlRMTVNTUAABAAAAl4II4gAAAAAAAAAAAAAAAAAAAAAKAGFKAAAADw=="}


class ntlm(BaseModule):
    watched_events = ["URL", "HTTP_RESPONSE"]
    produced_events = ["FINDING", "DNS_NAME"]
    flags = ["active", "safe", "web-basic"]
    meta = {"description": "Watch for HTTP endpoints that support NTLM authentication"}
    options = {"max_threads": 10, "try_all": False}
    options_desc = {"max_threads": "Maximum concurrent requests", "try_all": "Try every NTLM endpoint"}

    in_scope_only = True

    def setup(self):
        self.processed = set()
        self.processed_lock = Lock()
        self.found = set()
        self.try_all = self.config.get("try_all", False)
        return True

    def handle_event(self, event):
        found_hash = hash(f"{event.host}:{event.port}")
        if found_hash not in self.found:
            result_FQDN, request_url = self.handle_url(event)
            if result_FQDN and request_url:
                self.found.add(found_hash)
                self.emit_event(
                    {
                        "host": str(event.host),
                        "url": request_url,
                        "description": f"NTLM AUTH: {result_FQDN}",
                    },
                    "FINDING",
                    source=event,
                )
                self.emit_event(result_FQDN, "DNS_NAME", source=event)

    def filter_event(self, event):
        if self.try_all:
            return True
        if event.type == "HTTP_RESPONSE":
            if "www-authenticate" in event.data["header-dict"]:
                header_value = event.data["header-dict"]["www-authenticate"].lower()
                if "ntlm" in header_value or "negotiate" in header_value:
                    return True
        return False

    def handle_url(self, event):
        if event.type == "URL":
            urls = {
                event.data,
            }
        else:
            urls = {
                event.data["url"],
            }
        if self.try_all:
            for endpoint in ntlm_discovery_endpoints:
                urls.add(f"{event.parsed.scheme}://{event.parsed.netloc}/{endpoint}")

        futures = {}
        for url in urls:
            future = self.submit_task(self.check_ntlm, url)
            futures[future] = url

        for future in self.helpers.as_completed(futures):
            url = futures[future]
            try:
                result = future.result()
                if result:
                    for future in futures:
                        future.cancel()
                    return str(result["FQDN"]), url
            except RequestException as e:
                self.warning(str(e))

        return None, None

    def check_ntlm(self, test_url):
        url_hash = hash(test_url)

        with self.processed_lock:
            if url_hash in self.processed:
                return
            self.processed.add(url_hash)

        # use lower timeout value
        http_timeout = self.config.get("httpx_timeout", 5)
        r = self.helpers.request(
            test_url, headers=NTLM_test_header, raise_error=True, allow_redirects=False, timeout=http_timeout
        )
        ntlm_resp = r.headers.get("WWW-Authenticate", "")
        if ntlm_resp:
            ntlm_resp_b64 = max(ntlm_resp.split(","), key=lambda x: len(x)).split()[-1]
            try:
                ntlm_resp_decoded = self.helpers.ntlm.ntlmdecode(ntlm_resp_b64)
                if ntlm_resp_decoded:
                    return ntlm_resp_decoded
            except NTLMError as e:
                self.verbose(str(e))



import json
import subprocess
from bbot.modules.base import BaseModule


class naabu(BaseModule):
    watched_events = ["IP_ADDRESS", "DNS_NAME", "IP_RANGE"]
    produced_events = ["OPEN_TCP_PORT"]
    flags = ["active", "portscan", "aggressive"]
    meta = {"description": "Execute port scans with naabu"}
    options = {
        "ports": "",
        "top_ports": 100,
        "version": "2.1.1",
    }
    options_desc = {
        "ports": "ports to scan",
        "top_ports": "top ports to scan",
        "version": "naabu version",
    }
    max_event_handlers = 2
    batch_size = 100
    _priority = 2

    deps_ansible = [
        {
            "name": "install libpcap (Debian)",
            "package": {"name": "libpcap0.8", "state": "present"},
            "become": True,
            "when": """ansible_facts['os_family'] == 'Debian'""",
            "ignore_errors": True,
        },
        {
            "name": "install libpcap (others)",
            "package": {"name": "libpcap", "state": "present"},
            "become": True,
            "when": """ansible_facts['os_family'] != 'Debian'""",
            "ignore_errors": True,
        },
        {
            "name": "symlink libpcap",
            "file": {"src": "/usr/lib/libpcap.so", "dest": "#{BBOT_LIB}/libpcap.so.0.8", "state": "link"},
            "when": """ansible_facts['os_family'] != 'Debian'""",
            "ignore_errors": True,
        },
        {
            "name": "Download naabu",
            "unarchive": {
                "src": "https://github.com/projectdiscovery/naabu/releases/download/v#{BBOT_MODULES_NAABU_VERSION}/naabu_#{BBOT_MODULES_NAABU_VERSION}_#{BBOT_OS}_#{BBOT_CPU_ARCH}.zip",
                "include": "naabu",
                "dest": "#{BBOT_TOOLS}",
                "remote_src": True,
            },
        },
    ]

    def setup(self):
        self.helpers.depsinstaller.ensure_root(message="Naabu requires root privileges")
        return True

    def handle_batch(self, *events):
        _input = [str(e.data) for e in events]
        command = self.construct_command()
        for line in self.helpers.run_live(command, input=_input, stderr=subprocess.DEVNULL, sudo=True):
            try:
                j = json.loads(line)
            except Exception as e:
                self.debug(f'Error parsing line "{line}" as JSON: {e}')
                break
            host = j.get("host", j.get("ip"))
            port = j.get("port")

            source_event = None
            # check exact matches first
            for event in events:
                if host == str(event.host):
                    source_event = event
                    break
            # then make a broader check, for cidrs etc.
            if source_event is None:
                intermediary_event = None
                for event in events:
                    if host in event:
                        intermediary_event = event
                        break
                if intermediary_event is not None:
                    source_event = self.make_event(host, "IP_ADDRESS", source=intermediary_event)
                    if source_event:
                        self.emit_event(source_event)

            if source_event is None:
                self.warning(f'Failed to correlate source event for host "{host}"')
                continue

            self.emit_event(f"{host}:{port}", "OPEN_TCP_PORT", source=source_event)

    def construct_command(self):
        ports = self.config.get("ports", "")
        top_ports = self.config.get("top_ports", "")
        command = [
            "naabu",
            "-silent",
            "-json",
            # "-r",
            # self.helpers.resolver_file
        ]
        if ports:
            command += ["-p", ports]
        else:
            command += ["-top-ports", top_ports]
        return command

    def cleanup(self):
        resume_file = self.helpers.current_dir / "resume.cfg"
        resume_file.unlink(missing_ok=True)



import logging
from pathlib import Path
from bbot.modules.base import BaseModule


class BaseOutputModule(BaseModule):
    accept_dupes = True
    _type = "output"
    emit_graph_trail = True
    scope_distance_modifier = None
    _stats_exclude = True

    def _event_precheck(self, event):
        if event._omit:
            return False, "_omit is True"
        if event._force_output:
            return True, "_force_output is True"
        if event._internal:
            return False, "_internal is True"
        return super()._event_precheck(event)

    def _prep_output_dir(self, filename):
        self.output_file = self.config.get("output_file", "")
        if self.output_file:
            self.output_file = Path(self.output_file)
        else:
            self.output_file = self.scan.home / str(filename)
        self.helpers.mkdir(self.output_file.parent)
        self._file = None

    @property
    def file(self):
        if self._file is None:
            self._file = open(self.output_file, mode="a")
        return self._file

    @property
    def config(self):
        config = self.scan.config.get("output_modules", {}).get(self.name, {})
        if config is None:
            config = {}
        return config

    @property
    def log(self):
        if self._log is None:
            self._log = logging.getLogger(f"bbot.modules.output.{self.name}")
        return self._log



import re
import json
import requests

from bbot.modules.base import BaseModule
from bbot.core.helpers.misc import tldextract


class subdomain_hijack(BaseModule):
    flags = ["subdomain-hijack", "subdomain-enum", "cloud-enum", "safe", "active"]
    watched_events = ["DNS_NAME"]
    produced_events = ["FINDING"]
    meta = {"description": "Detect hijackable subdomains"}
    options = {
        "fingerprints": "https://raw.githubusercontent.com/blacklanternsecurity/can-i-take-over-xyz/master/fingerprints.json"
    }
    options_desc = {"fingerprints": "URL or path to fingerprints.json"}
    scope_distance_modifier = 2
    max_event_handlers = 5

    def setup(self):
        fingerprints_url = self.config.get("fingerprints")
        fingerprints_file = self.helpers.wordlist(fingerprints_url)
        with open(fingerprints_file) as f:
            fingerprints = json.load(f)
        self.fingerprints = []
        for f in fingerprints:
            try:
                f = Fingerprint(f)
            except Exception as e:
                self.warning(f"Error instantiating fingerprint: {e}")
                continue
            if not (f.domains and f.vulnerable and f.fingerprint and f.cicd_pass):
                self.debug(f"Skipping fingerprint: {f}")
                continue
            self.debug(f"Processed fingerprint: {f}")
            self.fingerprints.append(f)
        if not self.fingerprints:
            return None, "No valid fingerprints"
        self.debug(f"Successfully processed {len(self.fingerprints):,} fingerprints")
        return True

    def handle_event(self, event):
        hijackable, reason = self.check_subdomain(event)
        if hijackable:
            source_hosts = []
            e = event
            while 1:
                host = getattr(e, "host", "")
                if host:
                    if e not in source_hosts:
                        source_hosts.append(e)
                    e = e.source
                else:
                    break

            url = f"https://{event.host}"
            description = f'Hijackable Subdomain "{event.data}": {reason}'
            source_hosts = source_hosts[::-1]
            if source_hosts:
                source_hosts_str = str(source_hosts[0].host)
                for e in source_hosts[1:]:
                    source_hosts_str += f" -[{e.module.name}]-> {e.host}"
                description += f" ({source_hosts_str})"
            self.emit_event({"host": event.host, "url": url, "description": description}, "FINDING", source=event)
        else:
            self.debug(reason)

    def check_subdomain(self, event):
        for f in self.fingerprints:
            for domain in f.domains:
                self_matches = self.helpers.host_in_host(event.data, domain)
                child_matches = any(self.helpers.host_in_host(domain, h) for h in event.resolved_hosts)
                if self_matches or child_matches:
                    for scheme in ("https", "http"):
                        # first, try base request
                        url = f"{scheme}://{event.data}"
                        match, reason = self._verify_fingerprint(f, url)
                        if match:
                            return match, reason
                        # next, try {random_domain} -[DNS]-> domain
                        url = f"{scheme}://{domain}"
                        headers = {"Host": event.data}
                        match, reason = self._verify_fingerprint(f, url, headers=headers)
                        if match:
                            return match, reason
        return False, f'Subdomain "{event.data}" not hijackable'

    def _verify_fingerprint(self, fingerprint, *args, **kwargs):
        kwargs["raise_error"] = True
        if fingerprint.http_status is not None:
            kwargs["allow_redirects"] = False
        try:
            r = self.helpers.request(*args, **kwargs)
            if fingerprint.http_status is not None and r.status_code == fingerprint.http_status:
                return True, f"HTTP status == {fingerprint.http_status}"
            text = getattr(r, "text", "")
            if (
                not fingerprint.nxdomain
                and not fingerprint.http_status
                and fingerprint.fingerprint_regex.findall(text)
            ):
                return True, "Fingerprint match"
        except requests.exceptions.RequestException as e:
            if fingerprint.nxdomain and "Name or service not known" in str(e):
                return True, f"NXDOMAIN"
        return False, "No match"


class Fingerprint:
    def __init__(self, fingerprint):
        assert isinstance(fingerprint, dict), "fingerprint must be a dictionary"
        self.engine = fingerprint.get("service")
        self.cnames = fingerprint.get("cname", [])
        self.domains = list(set([tldextract(c).registered_domain for c in self.cnames]))
        self.http_status = fingerprint.get("http_status", None)
        self.nxdomain = fingerprint.get("nxdomain", False)
        self.vulnerable = fingerprint.get("vulnerable", False)
        self.fingerprint = fingerprint.get("fingerprint", "")
        self.cicd_pass = fingerprint.get("cicd_pass", False)
        try:
            self.fingerprint_regex = re.compile(self.fingerprint, re.MULTILINE)
        except re.error:
            self.fingerprint_regex = re.compile(re.escape(self.fingerprint), re.MULTILINE)

    def __str__(self):
        return f"{self.engine}: {self.fingerprint} (cnames: {self.cnames}, vulnerable: {self.vulnerable}, cicd_pass: {self.cicd_pass})"



import re
import requests_mock

from ..bbot_fixtures import *


def test_modules_basic(patch_commands, patch_ansible, scan, helpers, events, bbot_config, bbot_scanner):
    fallback_nameservers = scan.helpers.temp_dir / "nameservers.txt"
    with open(fallback_nameservers, "w") as f:
        f.write("8.8.8.8\n")

    with requests_mock.Mocker() as m:
        for http_method in ("GET", "CONNECT", "HEAD", "POST", "PUT", "TRACE", "DEBUG", "PATCH", "DELETE", "OPTIONS"):
            m.request(http_method, re.compile(r".*"), text='{"test": "test"}')

        # event filtering
        from bbot.modules.base import BaseModule
        from bbot.modules.output.base import BaseOutputModule
        from bbot.modules.report.base import BaseReportModule
        from bbot.modules.internal.base import BaseInternalModule

        # output module specific event filtering tests
        base_output_module = BaseOutputModule(scan)
        base_output_module.watched_events = ["IP_ADDRESS"]
        localhost = scan.make_event("127.0.0.1", source=scan.root_event)
        assert base_output_module._event_precheck(localhost)[0] == True
        localhost._internal = True
        assert base_output_module._event_precheck(localhost)[0] == False
        localhost._force_output = True
        assert base_output_module._event_precheck(localhost)[0] == True
        localhost._omit = True
        assert base_output_module._event_precheck(localhost)[0] == False

        # common event filtering tests
        for module_class in (BaseModule, BaseOutputModule, BaseReportModule, BaseInternalModule):
            base_module = module_class(scan)
            localhost2 = scan.make_event("127.0.0.2", source=events.subdomain)
            localhost2.make_in_scope()
            # base cases
            base_module._watched_events = None
            base_module.watched_events = ["*"]
            assert base_module._event_precheck(events.emoji)[0] == True
            base_module._watched_events = None
            base_module.watched_events = ["IP_ADDRESS"]
            assert base_module._event_precheck(events.ipv4)[0] == True
            assert base_module._event_precheck(events.domain)[0] == False
            assert base_module._event_precheck(events.localhost)[0] == True
            assert base_module._event_precheck(localhost2)[0] == True
            # target only
            base_module.target_only = True
            assert base_module._event_precheck(localhost2)[0] == False
            localhost2.tags.add("target")
            assert base_module._event_precheck(localhost2)[0] == True
            base_module.target_only = False
            # special case for IPs and ranges
            base_module.watched_events = ["IP_ADDRESS", "IP_RANGE"]
            ip_range = scan.make_event("127.0.0.0/24", dummy=True)
            localhost4 = scan.make_event("127.0.0.1", source=ip_range)
            localhost4.make_in_scope()
            localhost4.module = "plumbus"
            assert base_module._event_precheck(localhost4)[0] == True
            localhost4.module = "speculate"
            assert base_module._event_precheck(localhost4)[0] == False

            # in scope only
            localhost3 = scan.make_event("127.0.0.2", source=events.subdomain)
            base_module.in_scope_only = True
            assert base_module._event_postcheck(events.localhost)[0] == True
            assert base_module._event_postcheck(localhost3)[0] == False
            base_module.in_scope_only = False
            # scope distance
            base_module.scope_distance_modifier = 0
            localhost2._scope_distance = 0
            assert base_module._event_postcheck(localhost2)[0] == True
            localhost2._scope_distance = 1
            assert base_module._event_postcheck(localhost2)[0] == True
            localhost2._scope_distance = 2
            assert base_module._event_postcheck(localhost2)[0] == False
            localhost2._scope_distance = -1
            assert base_module._event_postcheck(localhost2)[0] == False
            base_module.scope_distance_modifier = -1

        base_output_module = BaseOutputModule(scan)
        base_output_module.watched_events = ["IP_ADDRESS"]

        scan2 = bbot_scanner(
            modules=list(set(available_modules + available_internal_modules)),
            output_modules=list(available_output_modules),
            config=bbot_config,
        )
        scan2.helpers.dns.fallback_nameservers_file = fallback_nameservers
        patch_commands(scan2)
        patch_ansible(scan2)
        scan2.load_modules()
        scan2.status = "RUNNING"

        # attributes, descriptions, etc.
        for module_name, module in scan2.modules.items():
            # flags
            assert module._type in ("internal", "output", "scan")

        # module preloading
        all_preloaded = module_loader.preloaded()
        assert "massdns" in all_preloaded
        assert "DNS_NAME" in all_preloaded["massdns"]["watched_events"]
        assert "DNS_NAME" in all_preloaded["massdns"]["produced_events"]
        assert "subdomain-enum" in all_preloaded["massdns"]["flags"]
        assert "wordlist" in all_preloaded["massdns"]["config"]
        assert type(all_preloaded["massdns"]["config"]["max_resolvers"]) == int
        assert all_preloaded["sslcert"]["deps"]["pip"]
        assert all_preloaded["sslcert"]["deps"]["apt"]
        assert all_preloaded["massdns"]["deps"]["ansible"]

        for module_name, preloaded in all_preloaded.items():
            # either active or passive and never both
            flags = preloaded.get("flags", [])
            if preloaded["type"] == "scan":
                assert ("active" in flags and not "passive" in flags) or (
                    not "active" in flags and "passive" in flags
                ), f'module "{module_name}" must have either "active" or "passive" flag'
                assert preloaded.get("meta", {}).get("description", ""), f"{module_name} must have a description"

            # attribute checks
            watched_events = preloaded.get("watched_events")
            produced_events = preloaded.get("produced_events")

            assert type(watched_events) == list
            assert type(produced_events) == list
            assert watched_events, f"{module_name}.watched_events must not be empty"
            assert type(watched_events) == list, f"{module_name}.watched_events must be of type list"
            assert type(produced_events) == list, f"{module_name}.produced_events must be of type list"
            assert all(
                [type(t) == str for t in watched_events]
            ), f"{module_name}.watched_events entries must be of type string"
            assert all(
                [type(t) == str for t in produced_events]
            ), f"{module_name}.produced_events entries must be of type string"

            assert type(preloaded.get("deps_pip", [])) == list, f"{module_name}.deps_pipe must be of type list"
            assert type(preloaded.get("deps_apt", [])) == list, f"{module_name}.deps_apt must be of type list"
            assert type(preloaded.get("deps_shell", [])) == list, f"{module_name}.deps_shell must be of type list"
            assert type(preloaded.get("config", None)) == dict, f"{module_name}.options must be of type list"
            assert (
                type(preloaded.get("options_desc", None)) == dict
            ), f"{module_name}.options_desc must be of type list"
            # options must have descriptions
            assert set(preloaded.get("config", {})) == set(
                preloaded.get("options_desc", {})
            ), f"{module_name}.options do not match options_desc"
            # descriptions most not be blank
            assert all(
                o for o in preloaded.get("options_desc", {}).values()
            ), f"{module_name}.options_desc descriptions must not be blank"

        # setups
        futures = {}
        for module_name, module in scan2.modules.items():
            log.info(f"Testing {module_name}.setup()")
            future = scan2._thread_pool.submit(module.setup)
            futures[future] = module
        for future in helpers.as_completed(futures):
            module = futures[future]
            result = future.result()
            if type(result) == tuple:
                assert len(result) == 2, f"if tuple, {module.name}.setup() return value must have length of 2"
                status, msg = result
                assert status in (
                    True,
                    False,
                    None,
                ), f"if tuple, the first element of {module.name}.setup()'s return value must be either True, False, or None"
                assert (
                    type(msg) == str
                ), f"if tuple, the second element of {module.name}.setup()'s return value must be a message of type str"
            else:
                assert result in (
                    True,
                    False,
                    None,
                ), f"{module.name}.setup() must return a status of either True, False, or None"
            if result == False:
                module.set_error_state()

        futures.clear()

        # handle_event / handle_batch
        futures = {}
        for module_name, module in scan2.modules.items():
            module.emit_event = lambda *args, **kwargs: None
            module._filter = lambda *args, **kwargs: True, ""
            events_to_submit = [e for e in events.all if e.type in module.watched_events]
            if module.batch_size > 1:
                log.info(f"Testing {module_name}.handle_batch()")
                future = scan2._thread_pool.submit(module.handle_batch, *events_to_submit)
                futures[future] = module
            else:
                for e in events_to_submit:
                    log.info(f"Testing {module_name}.handle_event()")
                    future = scan2._thread_pool.submit(module.handle_event, e)
                    futures[future] = module
        for future in helpers.as_completed(futures):
            try:
                assert future.result() == None
            except Exception as e:
                import traceback

                module = futures[future]
                assert module.errored == True, f'Error in module "{module}": {e}\n{traceback.format_exc()}'
        futures.clear()

        # finishes
        futures = {}
        for module_name, module in scan2.modules.items():
            log.info(f"Testing {module_name}.finish()")
            future = scan2._thread_pool.submit(module.finish)
            futures[future] = module
        for future in helpers.as_completed(futures):
            assert future.result() == None
        futures.clear()

        # cleanups
        futures = {}
        for module_name, module in scan2.modules.items():
            log.info(f"Testing {module_name}.cleanup()")
            future = scan2._thread_pool.submit(module.cleanup)
            futures[future] = module
        for future in helpers.as_completed(futures):
            assert future.result() == None
        futures.clear()

        # event filters
        for module_name, module in scan2.modules.items():
            log.info(f"Testing {module_name}.filter_event()")
            assert module.filter_event(events.emoji) in (
                True,
                False,
            ), f"{module_name}.filter_event() must return either True or False"



from ..bbot_fixtures import *


def test_cli(monkeypatch, bbot_config):
    from bbot import cli

    monkeypatch.setattr(sys, "exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(os, "_exit", lambda *args, **kwargs: True)
    monkeypatch.setattr(cli, "config", bbot_config)

    old_sys_argv = sys.argv

    # show version
    monkeypatch.setattr("sys.argv", ["bbot", "--version"])
    cli.main()

    # show current config
    monkeypatch.setattr("sys.argv", ["bbot", "-y", "--current-config"])
    cli.main()

    # list modules
    monkeypatch.setattr("sys.argv", ["bbot", "-l"])
    cli.main()

    home_dir = Path(bbot_config["home"])
    monkeypatch.setattr(
        sys,
        "argv",
        ["bbot", "-y", "-t", "www.example.com", "-om", "human", "csv", "json", "-n", "test_scan"],
    )
    cli.main()

    # unpatch sys.argv
    monkeypatch.setattr("sys.argv", old_sys_argv)

    scan_home = home_dir / "scans" / "test_scan"
    assert (scan_home / "wordcloud.tsv").is_file()
    assert (scan_home / "output.txt").is_file()
    assert (scan_home / "output.csv").is_file()
    assert (scan_home / "output.json").is_file()
    with open(scan_home / "output.csv") as f:
        lines = f.readlines()
        assert lines[0] == "Event type,Event data,IP Address,Source Module,Scope Distance,Event Tags\n"
        assert len(lines) > 1



from .base import BaseCloudProvider


class AWS(BaseCloudProvider):
    domains = [
        "amazon-dss.com",
        "amazonaws.com",
        "amazonaws.com.cn",
        "amazonaws.org",
        "amazonses.com",
        "amazonwebservices.com",
        "aws",
        "aws.a2z.com",
        "aws.amazon.com",
        "aws.dev",
        "awsstatic.com",
        "elasticbeanstalk.com",
    ]
    bucket_name_regex = r"[a-z0-9_][a-z0-9-\.]{1,61}[a-z0-9]"
    regexes = {
        "STORAGE_BUCKET": [r"(%[a-f0-9]{2})?(" + bucket_name_regex + r")\.(s3-?(?:[a-z0-9-]*\.){1,2}amazonaws\.com)"]
    }
